import logging
import re
from pathlib import Path

import ruamel.yaml

from yamlex.folder_structure import ensure_folder_structure
from yamlex.datasource_names import DATASOURCE_NAMES


logger = logging.getLogger(__name__)
parser = ruamel.yaml.YAML()


def split_yaml(extension_yaml_dir_path: Path) -> None:
    """Decompose a YAML file into multiple files."""
    extension_yaml_file_path = extension_yaml_dir_path / "extension.yaml"

    logger.info(f"Decomposing the central YAML file into parts: {extension_yaml_file_path}")
    with open(extension_yaml_file_path, "r") as extension_yaml_file:
        data = parser.load(extension_yaml_file)

    # Detect the datasource name
    datasource_name: str | None = None
    for name in DATASOURCE_NAMES:
        if name in data:
            datasource_name = name
            break

    logger.info(f"Datasource name detected: {datasource_name}")

    # Ensure the folder structure is in place
    ensure_folder_structure(extension_yaml_dir_path, datasource_name)
    logger.info(f"Folder structure ensured at: {extension_yaml_dir_path}")

    # Process datasource groups
    logger.info(f"Processing datasource groups...")
    if datasource_name and datasource_name != "python":
        datasource_dir_path = extension_yaml_dir_path / datasource_name

        items: list[dict] = data[datasource_name]
        for item in items:
            if datasource_name == "processes":
                item_name: str = item.get("name")
            else:
                item_name: str = item.get("group")

            if not item_name:
                logger.error(f"Item does not have a name")
                continue

            # Convert group name to a file name with only lowercase and underscores
            item_name_id = re.sub(r"[^a-z0-9_]", "_", item_name.lower())
            item_file_name = f"{item_name_id}.yaml"

            # Check whether such file already exists within the datasource directory
            search_results = list(datasource_dir_path.glob(f"**/{item_file_name}"))
            if search_results:
                logger.debug(f"Item file already exists: {search_results[0]}")
                continue

            # Extract group into a separate file
            item_file_path = datasource_dir_path / item_file_name
            with open(item_file_path, "w") as item_file:
                parser.indent(mapping=2, sequence=4, offset=2)
                parser.dump(item, item_file)
                logger.info(f"Item extracted: {item_file_path}")

    # Process metrics
    logger.info(f"Processing metrics...")
    metrics_dir_path = extension_yaml_dir_path / "metrics"
    if "metrics" in data:
        metrics: list[dict] = data["metrics"]
        for metric in metrics:
            entity_type = metric.get("key")
            if not entity_type:
                logger.error(f"Metric does not have a key: {metric}")
                continue

            metric_file_name = f"{entity_type}.yaml"

            # Check whether such file already exists within the metrics directory
            search_results = list(metrics_dir_path.glob(f"**/{metric_file_name}"))
            if search_results:
                logger.debug(f"Metric file already exists: {search_results[0]}")
                continue

            # Extract metric into a separate file
            metric_file_path = metrics_dir_path / metric_file_name
            with open(metric_file_path, "w") as metric_file:
                parser.indent(mapping=2, sequence=4, offset=2)
                parser.dump(metric, metric_file)
                logger.info(f"Metric extracted: {metric_file_path}")

    # Process screens
    logger.info(f"Processing screens...")
    screens_dir_path = extension_yaml_dir_path / "screens"
    if "screens" in data:
        screens: list[dict] = data["screens"]
        for screen in screens:
            entity_type: str = screen.get("entityType")
            if not entity_type:
                logger.error(f"Screen does not have an entity_type")
                continue

            entity_type_id = re.sub(r"[^a-z0-9_]", "_", entity_type.lower())
            screen_file_name = f"{entity_type_id}.yaml"

            # Check whether such file already exists within the screens directory
            search_results = list(screens_dir_path.glob(f"**/{screen_file_name}"))
            if search_results:
                logger.debug(f"Screen file already exists: {search_results[0]}")
                continue

            # Extract screen into a separate file
            screen_file_path = screens_dir_path / screen_file_name
            with open(screen_file_path, "w") as screen_file:
                parser.indent(mapping=2, sequence=4, offset=2)
                parser.dump(screen, screen_file)
                logger.info(f"Screen extracted: {screen_file_path}")

    # Process topology
    logger.info(f"Processing topology...")
    topology_dir_path = extension_yaml_dir_path / "topology"
    if "topology" in data:
        topology: dict = data["topology"]
        if "types" in topology:
            logger.info(f"Processing topology types...")
            types: list[dict] = topology["types"]
            types_dir_path = topology_dir_path / "types"
            for type_ in types:
                type_name = type_.get("name")
                if not type_name:
                    logger.error(f"Type does not have a name: {type_}")
                    continue

                type_name_id = re.sub(r"[^a-z0-9_]", "_", type_name.lower())
                type_file_name = f"{type_name_id}.yaml"

                # Check whether such file already exists within the types directory
                search_results = list(types_dir_path.glob(f"**/{type_file_name}"))
                if search_results:
                    logger.debug(f"Type file already exists: {search_results[0]}")
                    continue

                # Extract type into a separate file
                type_file_path = types_dir_path / type_file_name
                with open(type_file_path, "w") as type_file:
                    parser.indent(mapping=2, sequence=4, offset=2)
                    parser.dump(type_, type_file)
                    logger.info(f"Type extracted: {type_file_path}")

        if "relationships" in topology:
            logger.info(f"Processing topology relationships...")
            relationships: list[dict] = topology["relationships"]
            relationships_dir_path = topology_dir_path / "relationships"
            for relationship in relationships:
                if (
                    "fromType" not in relationship
                    or "toType" not in relationship
                    or "typeOfRelation" not in relationship
                ):
                    logger.error(f"Relationship does not have required fields: {relationship}")
                    continue

                relationship_from_id = re.sub(r"[^a-z0-9_]", "_", relationship["fromType"].lower())
                relationship_to_id = re.sub(r"[^a-z0-9_]", "_", relationship["toType"].lower())
                relationship_type_id = re.sub(r"[^a-z0-9_]", "_", relationship["typeOfRelation"].lower())

                relationship_name_id = f"{relationship_from_id}_{relationship_type_id}_{relationship_to_id}"
                relationship_file_name = f"{relationship_name_id}.yaml"

                # Check whether such file already exists within the relationships directory
                search_results = list(relationships_dir_path.glob(f"**/{relationship_file_name}"))
                if search_results:
                    logger.debug(f"Relationship file already exists: {search_results[0]}")
                    continue

                # Extract relationship into a separate file
                relationship_file_path = relationships_dir_path / relationship_file_name
                with open(relationship_file_path, "w") as relationship_file:
                    parser.indent(mapping=2, sequence=4, offset=2)
                    parser.dump(relationship, relationship_file)
                    logger.info(f"Relationship extracted: {relationship_file_path}")



import logging
from pathlib import Path
from typing import Optional

import typer
from typing_extensions import Annotated

from yamlex.schema_mapper import update_vscode_settings_json
from yamlex.folder_structure import ensure_folder_structure
from yamlex.splitter import split_yaml
from yamlex.joiner import join_yaml


logging.basicConfig(level=logging.INFO, format="%(message)s")

logger = logging.getLogger(__name__)
app = typer.Typer(
    name="yamlex",
    rich_markup_mode="rich",
)

yaml_option = Annotated[
    Optional[Path],
    typer.Option(
        "--yaml",
        "-y",
        help=(
            "The directory containing the YAML files. "
            "Defaults to: 'src/extension/' or 'extension/'"
        ),
        show_default=False,
    ),
]
json_option = Annotated[
    Path,
    typer.Option(
        "--json",
        "-j",
        help="Path to the directory with JSON schema files.",
        exists=True,
        readable=True,
    ),
]
settings_option = Annotated[
    Path,
    typer.Option(
        "--settings",
        "-s",
        help="Path to the VS Code settings.json file.",
    ),
]
output_option = Annotated[
    Optional[Path],
    typer.Option(
        "--output",
        "-o",
        help="Path to output extension.yaml file.",
        show_default=False,
    ),
]
force_option = Annotated[
    bool,
    typer.Option(
        "--force",
        "-f",
        help="Overwrite the destination file if it was created manually.",
    ),
]
cwd_option = Annotated[
    Optional[Path],
    typer.Option(
        "--cwd",
        help="The current working directory to use.",
        exists=True,
        readable=True,
        hidden=True,
    ),
]
verbose_option = Annotated[
    bool,
    typer.Option(
        "--verbose",
        "-v",
        help="Enable verbose output.",
    ),
]
quiet_option = Annotated[
    bool,
    typer.Option(
        "--quiet",
        "-q",
        help="Disable any informational output. Only errors.",
    ),
]


def adjust_logging(verbose: bool = False, quiet: bool = False) -> None:
    if quiet:
        logging.root.setLevel(logging.ERROR)
    elif verbose:
        logging.root.setLevel(logging.DEBUG)


def get_default_extension_dir_path() -> Path:
    candidates = [
        Path("extension"),
        Path("src/extension"),
    ]

    # Find the first suitable candidate
    for candidate in candidates:
        # Check the candidate directory exists
        if candidate.exists() and candidate.is_dir():
            return candidate
        continue

    # Nothing suitable was found
    logger.error("No extension directory found")
    exit(1)


@app.command(
    help=(
        "Enable VS Code to validate YAML schema of individual parts "
        "by mapping each of them to proper JSON schema in local "
        "VS Code settings file."
    ),
    name="map",
)
def map_(
    json: json_option = Path("schema"),
    settings: settings_option = Path(".vscode/settings.json"),
    yaml: yaml_option = None,
    cwd: cwd_option = Path("."),
    verbose: verbose_option = False,
    quiet: quiet_option = False,
):
    """Map YAML files to JSON schema."""
    adjust_logging(verbose, quiet)

    logger.info("Mapping YAML files to JSON schema...")
    logger.debug(f"Current working directory: {cwd}")

    json_schemas_dir_path = json
    if not json_schemas_dir_path.is_absolute():
        json_schemas_dir_path = cwd / json_schemas_dir_path
    logger.debug(f"JSON schema files directory: {json}")

    vscode_settings_json_file_path = settings
    if not vscode_settings_json_file_path.is_absolute():
        vscode_settings_json_file_path = cwd / vscode_settings_json_file_path
    logger.debug(f"VS Code settings.json file: {settings}")

    extension_yaml_dir_path = yaml or get_default_extension_dir_path()
    if not extension_yaml_dir_path.is_absolute():
        extension_yaml_dir_path = cwd / extension_yaml_dir_path
    logger.debug(f"Extension YAML files directory: {yaml}")

    update_vscode_settings_json(
        json_schemas_dir_path,
        vscode_settings_json_file_path,
        extension_yaml_dir_path,
    )


@app.command(help="Ensure the folder structure to store YAML files is in place.")
def structure(
    datasource_name: Annotated[str, typer.Argument(..., help="The name of the datasource.")],
    yaml: yaml_option = None,
    cwd: cwd_option = Path("."),
    verbose: verbose_option = False,
    quiet: quiet_option = False,
):
    """Ensure the folder structure is in place."""
    adjust_logging(verbose, quiet)

    logger.info("Ensuring the folder structure is in place...")

    extension_yaml_dir_path = yaml or get_default_extension_dir_path()
    if not extension_yaml_dir_path.is_absolute():
        extension_yaml_dir_path = cwd / extension_yaml_dir_path
    logger.debug(f"Extension YAML files directory: {extension_yaml_dir_path}")

    ensure_folder_structure(extension_yaml_dir_path, datasource_name)


@app.command(help="Split central YAML file into parts.")
def split(
    yaml: yaml_option = None,
    cwd: cwd_option = Path("."),
    verbose: verbose_option = False,
    quiet: quiet_option = False,
):
    """Decompose the central YAML file into parts."""
    adjust_logging(verbose, quiet)

    logger.info("Decomposing the central YAML file into parts...")

    extension_yaml_dir_path = yaml or get_default_extension_dir_path()
    if not extension_yaml_dir_path.is_absolute():
        extension_yaml_dir_path = cwd / extension_yaml_dir_path
    logger.debug(f"Extension YAML files directory: {yaml}")

    split_yaml(extension_yaml_dir_path)


@app.command(help="Join YAML files into a single file.")
def join(
    yaml: yaml_option = None,
    output: output_option = None,
    force: force_option = False,
    cwd: cwd_option = Path("."),
    verbose: verbose_option = False,
    quiet: quiet_option = False,
):
    """Run the application."""
    adjust_logging(verbose, quiet)

    extension_yaml_dir_path = yaml or get_default_extension_dir_path()
    if not extension_yaml_dir_path.is_absolute():
        extension_yaml_dir_path = cwd / extension_yaml_dir_path
    logger.debug(f"Extension YAML files directory: {yaml}")

    output_file_path = output or extension_yaml_dir_path / "extension.yaml"
    if not output_file_path.is_absolute():
        output_file_path = cwd / output_file_path
    logger.debug(f"Output file: {output}")

    join_yaml(extension_yaml_dir_path, output_file_path, force)



from pathlib import Path


def ensure_folder_structure(extension_yaml_dir_path: Path, datasource_name: str | None = None):
    """Ensure the folder structure is in place."""
    # metrics
    metrics_dir_path = extension_yaml_dir_path / "metrics"
    metrics_dir_path.mkdir(parents=True, exist_ok=True)

    # datasource
    if datasource_name and datasource_name != "python":
        datasource_dir_path = extension_yaml_dir_path / datasource_name
        datasource_dir_path.mkdir(parents=True, exist_ok=True)

    # topology
    topology_dir_path = extension_yaml_dir_path / "topology"
    topology_dir_path.mkdir(parents=True, exist_ok=True)

    # topology types
    topology_types_dir_path = topology_dir_path / "types"
    topology_types_dir_path.mkdir(parents=True, exist_ok=True)

    # topology relationships
    topology_relationships_dir_path = topology_dir_path / "relationships"
    topology_relationships_dir_path.mkdir(parents=True, exist_ok=True)

    # screens
    screens_dir_path = extension_yaml_dir_path / "screens"
    screens_dir_path.mkdir(parents=True, exist_ok=True)



DATASOURCE_NAMES = [
    "gcp",
    "jmx",
    "python",
    "snmp",
    "snmptraps",
    "sqlDb2",
    "sqlHana",
    "sqlMySql",
    "sqlOracle",
    "sqlPostgres",
    "sqlServer",
    "sqlSnowflake",
    "wmi",
]



import logging
from datetime import datetime
from pathlib import Path
from io import StringIO

import ruamel.yaml


logger = logging.getLogger(__name__)
parser = ruamel.yaml.YAML()


def join_yaml(
    extension_yaml_dir_path: Path,
    output_file_path: Path,
    force: bool = False,
) -> None:
    """Assemble YAML files into a single file."""
    logger.info("Assembling YAML files into a single file...")

    # Make sure extension.base.yaml file exists
    extension_base_yaml_file_path = extension_yaml_dir_path / "extension.base.yaml"
    if not (extension_base_yaml_file_path.exists() and extension_base_yaml_file_path.is_file()):
        logger.error(f"The {extension_base_yaml_file_path} file was not found")
        exit(1)

    with open(extension_base_yaml_file_path, "r") as extension_base_yaml_file:
        extension = parser.load(extension_base_yaml_file)

    # Assemble metrics
    assemble_metrics(extension, extension_yaml_dir_path)

    # Assemble datasource
    assemble_datasource(extension, "sqlPostgres", extension_yaml_dir_path)

    # Assemble topology
    assemble_topology(extension, extension_yaml_dir_path)

    # Assemble screens
    assemble_screens(extension, extension_yaml_dir_path)

    # Check if the output file exists and is was created manually
    if output_file_path.exists():
        with open(output_file_path, "r") as output_file:
            existing_output = output_file.read()
            if "generated by yamlex" not in existing_output and not force:
                logger.error(f"The {output_file_path} file was created manually. Use --force to overwrite it.")
                exit(2)

    # Convert dict to YAML. Dump to string first to add a comment
    stream = StringIO()
    parser.indent(mapping=2, sequence=4, offset=2)
    parser.dump(extension, stream)
    extension_yaml_text = stream.getvalue()

    # Write to output file
    with open(output_file_path, "w") as output_file:
        # Write a comment to indicate that the file was automatically generated
        comment = (
            "# This file was generated by yamlex\n"
            f"# at: {datetime.now().isoformat()}\n"
            "\n"
        )
        output_file.write(comment)
        output_file.write(extension_yaml_text)


def assemble_metrics(extension: dict, extension_yaml_dir_path: Path) -> None:
    metrics_dir_path = extension_yaml_dir_path / "metrics"
    metric_yaml_file_paths = list(metrics_dir_path.glob("**/*.yaml"))
    logger.info(f"Found {len(metric_yaml_file_paths)} metrics")

    if metric_yaml_file_paths:
        extension["metrics"] = []

    for metric_yaml_file_path in metric_yaml_file_paths:
        with open(metric_yaml_file_path, "r") as metric_yaml_file:
            metric = parser.load(metric_yaml_file)
            extension["metrics"].append(metric)


def assemble_datasource(extension: dict, name: str, extension_yaml_dir_path: Path) -> None:
    datasource_dir_path = extension_yaml_dir_path / name
    group_yaml_file_paths = list(datasource_dir_path.glob("*.yaml"))
    logger.info(f"Found {len(group_yaml_file_paths)} groups")

    if group_yaml_file_paths:
        extension[name] = []

    for group_yaml_file_path in group_yaml_file_paths:
        with open(group_yaml_file_path, "r") as group_yaml_file:
            group = parser.load(group_yaml_file)
            extension[name].append(group)


def assemble_topology(extension: dict, extension_yaml_dir_path: Path) -> None:
    types_dir_path = extension_yaml_dir_path / "topology" / "types"
    type_yaml_file_paths = list(types_dir_path.glob("*.yaml"))
    logger.info(f"Found {len(type_yaml_file_paths)} topology types")

    relationships_dir_path = extension_yaml_dir_path / "topology" / "relationships"
    relationship_yaml_file_paths = list(relationships_dir_path.glob("*.yaml"))
    logger.info(f"Found {len(relationship_yaml_file_paths)} topology relationships")

    if type_yaml_file_paths or relationship_yaml_file_paths:
        extension["topology"] = {}

    if type_yaml_file_paths:
        extension["topology"]["types"] = []

    for type_yaml_file_path in type_yaml_file_paths:
        with open(type_yaml_file_path, "r") as type_yaml_file:
            type_ = parser.load(type_yaml_file)
            extension["topology"]["types"].append(type_)

    if relationship_yaml_file_paths:
        extension["topology"]["relationships"] = []

    for relationship_yaml_file_path in relationship_yaml_file_paths:
        with open(relationship_yaml_file_path, "r") as relationship_yaml_file:
            relationship = parser.load(relationship_yaml_file)
            extension["topology"]["relationships"].append(relationship)


def assemble_screens(extension: dict, extension_yaml_dir_path: Path) -> None:
    screens_dir_path = extension_yaml_dir_path / "screens"
    screen_yaml_file_paths = list(screens_dir_path.glob("*.yaml"))
    logger.info(f"Found {len(screen_yaml_file_paths)} screens")

    if screen_yaml_file_paths:
        extension["screens"] = []

    for screen_yaml_file_path in screen_yaml_file_paths:
        with open(screen_yaml_file_path, "r") as screen_yaml_file:
            screen = parser.load(screen_yaml_file)
            extension["screens"].append(screen)



import json
import logging
import re
from pathlib import Path


logger = logging.getLogger(__name__)


def update_vscode_settings_json(
    json_schemas_dir_path: Path,
    vscode_settings_json_file_path: Path,
    extension_yaml_dir_path: Path,
):
    """Map YAML files to JSON schema."""
    # Check that directory with Extension JSON schema files exists and are valid
    check_json_schemas_dir_exists(json_schemas_dir_path)

    # Check that directory with Extension YAML files exists
    check_extension_yaml_dir_exists(extension_yaml_dir_path)

    # Convert each definition from extension.schema.json to a separate schema file
    extract_definitions_into_standalone_schemas(json_schemas_dir_path)

    # Read settings.json if it exists, otherwise create an empty dictionary
    project_dir_path = vscode_settings_json_file_path.parent.parent
    vscode_settings = read_vscode_settings_json_file(vscode_settings_json_file_path)
    
    # Check if yaml.schemas mapping field exists in settings.json
    if "yaml.schemas" not in vscode_settings:
        vscode_settings["yaml.schemas"] = {}

    # Map extension JSON schema definitions to extension YAML files
    map_json_schemas_to_yaml(
        vscode_settings,
        project_dir_path,
        json_schemas_dir_path,
        extension_yaml_dir_path,
    )

    # Write the updated settings.json file
    with open(vscode_settings_json_file_path, "w") as f:
        vscode_settings_file_text = json.dumps(vscode_settings, indent=2)
        f.write(vscode_settings_file_text)
        logger.info(f"Updated {vscode_settings_json_file_path} with new YAML schema mappings.")


def check_json_schemas_dir_exists(json_schemas_dir_path: Path):
    """Make sure that source directory with JSON schema exists and contains valid JSON schema files"""
    if not json_schemas_dir_path.exists():
        logger.error(f"{json_schemas_dir_path} directory with JSON schema files does not exist.")
        exit(1)
    
    if not (json_schemas_dir_path / "extension.schema.json").exists():
        logger.error(f"{json_schemas_dir_path} directory does not contain extension.schema.json file.")
        exit(1)
    

def check_extension_schema_json_file_exists(
    json_schemas_dir_path: Path,
    extension_schema_json_file_path: Path,
):
    """Make sure that extension.schema.json file exists in the source directory with JSON schema"""
    schema_names = list(json_schemas_dir_path.glob("*.json"))
    if extension_schema_json_file_path not in schema_names:
        logger.error(f"extension.schema.json not found in {json_schemas_dir_path}.")
        exit(1)

    if not extension_schema_json_file_path.exists():
        logger.error(f"extension.schema.json not found in {json_schemas_dir_path}.")
        exit(1)


def check_extension_yaml_dir_exists(extension_yaml_dir_path: Path):
    """Make sure files directory with YAML files exists"""
    if not extension_yaml_dir_path.exists():
        logger.error(f"{extension_yaml_dir_path} directory with extension YAML files does not exist.")
        exit(1)


def extract_definitions_into_standalone_schemas(json_schemas_dir_path: Path):
    """Open each schema file and extract definitions into standalone schema files"""
    # Get a list of all JSON schema files in the directory
    schema_file_paths = list(json_schemas_dir_path.glob("*.json"))
    for schema_file_path in schema_file_paths:
        relative_schema_file_path = schema_file_path.relative_to(json_schemas_dir_path)
        # Get stem two times to convert extension.schema.json to extension
        parent_schema_stem = Path(schema_file_path.stem).stem 
        logger.debug(f"Processing schema file: {relative_schema_file_path}")

        with open(schema_file_path, "r") as f:
            schema_json_text = f.read()
            schema = json.loads(schema_json_text)

            # Do not process already extracted schemas
            if not str(schema.get("$id")).startswith("http"):
                # Skip
                continue

            # Collect potential new schemas to process
            new_schemas_to_process: dict[str, dict] = {}

            if "definitions" in schema:
                for d, definition in schema["definitions"].items():
                    # Check if the definition is nested
                    if "enums" == d:
                        logger.debug(f"Nested enums found in {relative_schema_file_path}.")
                        for enum_name, enum_def in definition.items():
                            logger.debug(f"Extracting enum: {enum_name}")
                            new_schemas_to_process[f"enums/{enum_name}"] = enum_def
                    elif "types" == d:
                        for type_name, type_def in definition.items():
                            logger.debug(f"Extracting type: {type_name}")
                            new_schemas_to_process[f"types/{type_name}"] = type_def
                    elif (
                        definition.get("type") == "array"
                        and "items" in definition
                        and "$ref" not in definition["items"]
                    ):
                        logger.debug(f"Extracting array item definition: {d}")
                        new_schemas_to_process[d] = definition["items"]
                    else:
                        logger.debug(f"Extracting definition: {d}")
                        new_schemas_to_process[d] = definition

            if "properties" in schema:
                for p, property in schema["properties"].items():
                    if (
                        property.get("type") == "array"
                        and "items" in property
                        and "$ref" not in property["items"]
                    ):
                        logger.debug(f"Extracting array item property: {p}")
                        new_schemas_to_process[p] = property["items"]

            if (
                schema.get("type") == "array"
                and "items" in schema
                and "$ref" not in schema["items"]
            ):
                logger.debug(f"Extracting child item from array: {relative_schema_file_path}")
                new_schemas_to_process["object"] = schema["items"]

            for name, new_schema in new_schemas_to_process.items():
                valid_name = name.replace("/", ".")
                extract_single_definition_into_standalone_schema(
                    parent_schema_stem,
                    valid_name,
                    json_schemas_dir_path,
                    new_schema,
                )


def extract_single_definition_into_standalone_schema(
    parent_schema_stem: str,
    name: str,
    json_schemas_dir_path: Path,
    definition: dict,
):
    file_name = f"{parent_schema_stem}.{name}.schema.json"
    enrich_definition_for_standalone_schema(definition, file_name)

    # Write the enriched definition to a new schema file
    d_schema_file_path = json_schemas_dir_path / file_name
    with open(d_schema_file_path, "w") as d_schema_file:
        d_schema_text = json.dumps(definition, indent=2)

        # If extracted definition contains references to sibling definitions,
        # then replace them with new extracted schema file names
        # Regex to replace "#/definitions/{name}"" reference with "{name}.schema.json"

        # Find all references to sibling definitions using regex
        for m in re.finditer(r'"#/definitions/(.+)"', d_schema_text):
            full_match = m.group(0)
            logger.debug(f"Found reference to sibling definition: {full_match}")
            catch_replacement = m.group(1).replace("/", ".")
            full_replacement = f'"{parent_schema_stem}.{catch_replacement}.schema.json"'
            logger.debug(f'Will be replaced with: {full_replacement}')
            d_schema_text = d_schema_text.replace(full_match, full_replacement)

        # Save to file
        d_schema_file.write(d_schema_text)
        logger.info(f"Extracted new schema: {file_name}")


def enrich_definition_for_standalone_schema(definition: dict, file_name: str):
    definition["$schema"] = "http://json-schema.org/draft-07/schema#"
    definition["$id"] = file_name
    definition["description"] = f"Representation of {file_name}"
    definition["additionalProperties"] = False


def read_vscode_settings_json_file(vscode_settings_json_file_path: Path):
    """Read settings.json if it exists, otherwise create an empty dictionary."""
    if vscode_settings_json_file_path.exists():
        with open(vscode_settings_json_file_path, "r") as f:
            vscode_settings_file_text = f.read()
            vscode_settings = json.loads(vscode_settings_file_text)
    else:
        vscode_settings = {}
    return vscode_settings


def map_json_schemas_to_yaml(
    vscode_settings: dict,
    project_dir_path: Path,
    extension_schema_dir_path: Path,
    extension_yaml_dir_path: Path,
):
    """Associate extension.schema.json with extension.yaml and extension.base.yaml"""
    schemas = vscode_settings.get("yaml.schemas", {})

    # Delete extension.yaml mapping
    for k, v in schemas.items():
        if "extension.yaml" in str(v):
            del schemas[k]
            break

    # If project dir is just the current directory, then resolve it
    project_dir_path = project_dir_path.resolve()

    # Make sure the directories are relative to the project directory
    extension_schema_dir_path = extension_schema_dir_path.resolve().relative_to(project_dir_path)
    extension_yaml_dir_path = extension_yaml_dir_path.resolve().relative_to(project_dir_path)

    # ------------------------------
    # Generated extension assembly
    # ------------------------------

    schemas[(extension_schema_dir_path / "extension.schema.json").as_posix()] = [
        (extension_yaml_dir_path / "extension.yaml").as_posix(),
        (extension_yaml_dir_path / "extension.base.yaml").as_posix(),
    ]

    # ------------------------------
    # Metrics
    # ------------------------------

    schemas[(extension_schema_dir_path / "extension.metrics.schema.json").as_posix()] = [
        (extension_yaml_dir_path / "metrics" / "**/*.yaml").as_posix(),
    ]

    # ------------------------------
    # Datasources
    # ------------------------------

    # gcp
    schemas[(extension_schema_dir_path / "gcp.service.schema.json").as_posix()] = [
        (extension_yaml_dir_path / "gcp" / "*.yaml").as_posix(),
    ]

    # jmx
    schemas[(extension_schema_dir_path / "jmx.group.schema.json").as_posix()] = [
        (extension_yaml_dir_path / "jmx" / "*.yaml").as_posix(),
    ]

    # processes
    schemas[(extension_schema_dir_path / "processes.object.schema.json").as_posix()] = [
        (extension_yaml_dir_path / "processes" / "*.yaml").as_posix(),
    ]

    # prometheus
    schemas[(extension_schema_dir_path / "prometheus.object.schema.json").as_posix()] = [
        (extension_yaml_dir_path / "prometheus" / "*.yaml").as_posix(),
    ]

    # snmp
    schemas[(extension_schema_dir_path / "snmp.object.schema.json").as_posix()] = [
        (extension_yaml_dir_path / "snmp" / "*.yaml").as_posix(),
    ]

    # snmptraps
    schemas[(extension_schema_dir_path / "snmptraps.object.schema.json").as_posix()] = [
        (extension_yaml_dir_path / "snmptraps" / "*.yaml").as_posix(),
    ]

    # sql
    schemas[(extension_schema_dir_path / "sql.object.schema.json").as_posix()] = [
        (extension_yaml_dir_path / "sqlDb2" / "*.yaml").as_posix(),
        (extension_yaml_dir_path / "sqlHana" / "*.yaml").as_posix(),
        (extension_yaml_dir_path / "sqlMySql" / "*.yaml").as_posix(),
        (extension_yaml_dir_path / "sqlPostgres" / "*.yaml").as_posix(),
        (extension_yaml_dir_path / "sqlOracle" / "*.yaml").as_posix(),
        (extension_yaml_dir_path / "sqlServer" / "*.yaml").as_posix(),
        (extension_yaml_dir_path / "sqlSnowflake" / "*.yaml").as_posix(),
    ]

    # wmi
    schemas[(extension_schema_dir_path / "wmi.object.schema.json").as_posix()] = [
        (extension_yaml_dir_path / "wmi" / "*.yaml").as_posix(),
    ]

    # ------------------------------
    # Topology
    # ------------------------------

    schemas[(extension_schema_dir_path / "generic.types.object.schema.json").as_posix()] = [
        (extension_yaml_dir_path / "topology" / "types" / "**/*.yaml").as_posix(),
    ]

    schemas[(extension_schema_dir_path / "generic.relationships.object.schema.json").as_posix()] = [
        (extension_yaml_dir_path / "topology" / "relationships" / "**/*.yaml").as_posix(),
    ]
    
    # ------------------------------
    # Screens
    # ------------------------------

    schemas[(extension_schema_dir_path / "extension.screens.schema.json").as_posix()] = [
        (extension_yaml_dir_path / "screens" / "**/*.yaml").as_posix(),
    ]

    # Update yaml.schemas mapping in settings.json
    vscode_settings["yaml.schemas"] = schemas


